#include "liste.h"

Bool estVideListe(liste laListe){
    return laListe == NULL;
}


liste listeNouv(){
    return NULL;
}

Nat accesTete(liste laListe){
    //* Pré-condition : liste non vide.
    return laListe->valeur;
}

liste adjonctionEntete(liste laListe, Nat valeurAAjouter){
    liste noeud = MALLOC(strListe);
    if(noeud == NULL){
        fprintf(stderr, "Erreur : Allocation échouée!");
        exit(EXIT_FAILURE);
    }
    noeud->valeur = valeurAAjouter ;
    noeud->precedent = NULL;
    noeud->suivant = NULL;
    //* Cas simple (liste vide);
    if(estVideListe(laListe)){
        laListe = noeud;
    }
    //* Cas général ()
    else {
        noeud->suivant = laListe;
        laListe->precedent = noeud;
        laListe = noeud;
    }
    
    return laListe;
}

liste suppressionEnTete(liste laListe){
    //* Pré-condition : liste non vide
    //* Cas simple (Un seul élément)
    if(laListe->suivant == NULL){
        FREE(laListe);
        laListe = listeNouv();
    }
    //* Cas général
    else {
        liste noeudASupprimer = laListe;
        laListe->suivant->precedent = NULL;
        laListe = laListe->suivant ;
        FREE(noeudASupprimer);
    }

    return laListe;
}

Nat longueurListe(liste laListe){
    Nat longueur = 0 ;
    liste temp = laListe;

    while(temp != NULL){
        longueur++;
        temp = temp->suivant;
    }

    return longueur;
}

Bool appartient(liste laListe , Nat valeur){
    Bool trouve = false;
    liste temp = laListe;

    while(temp != NULL && !trouve){
        if(temp->valeur == valeur){
            trouve = true;
        }
        temp = temp->suivant;
    }

    return trouve;
}


//* Opération en I-ème position
Nat accesElementPositionI(liste laListe , Nat position){
    //* Pré-condition : position app [1, longueurListe];
    Nat longueur = longueurListe(laListe);
    liste temp = laListe;

    for( int i = 1 ; i < position ; i++){
        temp = temp->suivant;
    }

    return temp->valeur;
}

//* Retourne l'addresse du noeud avant ou NULL sinon
liste addresseNoeudPositionI(liste laListe, Nat position){
    //* Pré-condition : position >= 1;
    liste temp = laListe;
    liste addNoeudPosI ;
    Nat i = 1 ;
    while( i != position && temp != NULL){
        i++;
        temp = temp->suivant;
    }

    //* temp = addNoeudPosI || NULL;
    addNoeudPosI = temp ;

    return addNoeudPosI ;
}

liste insertionPositionI(liste laListe, Nat position, Nat valeurAInserer){
    liste noeud = MALLOC(strListe);

    if( noeud == NULL){
        fprintf(stderr, "Erreur : Allocation mémoire");
        exit(EXIT_FAILURE);
    }

    noeud->valeur = valeurAInserer;
    noeud->suivant = NULL;
    noeud->precedent = NULL;

    if(estVideListe(laListe) || position == 1){
        laListe = adjonctionEntete(laListe, valeurAInserer);
    }else {
        liste noeudPosI = addresseNoeudPositionI(laListe, position);
        if( noeudPosI != NULL){
            //* Il y a un noeud à la position I
            liste noeudAvant = noeudPosI->precedent ;
            liste noeudApres = noeudAvant->suivant ;

            noeud->suivant = noeudApres ;
            noeud->precedent = noeudAvant ;

            noeudAvant->suivant = noeud ;
            noeudApres->precedent = noeud ;
        }
    }

    return laListe;    
}


liste changementPositionI(liste laListe, Nat position, Nat nouvelleValeur){
    //* Pré-condition : position appartient [0, longueurListe]
    liste noeud = addresseNoeudPositionI(laListe, position);
    noeud->valeur = nouvelleValeur ;
    
    return laListe;
}

liste suppressionPositionI(liste laListe, Nat position){
    //* Pré-condition : position appartient [1 , longueurListe]

    //* Si c'est le 1 er élément.
    if(position == 1){
        laListe = suppressionEnTete(laListe);
    }else if(position == longueurListe(laListe)) {
    //* Si c'est le dernier élément.
        liste noeudASupprimer = addresseNoeudPositionI(laListe, position);
        liste noeudAvant = noeudASupprimer->precedent ;
        noeudAvant->suivant = NULL;
        FREE(noeudASupprimer);
    }
    else {
        liste noeudASupprimer = addresseNoeudPositionI(laListe, position);
        liste noeudAvant = noeudASupprimer->precedent ;
        liste noeudApres = noeudASupprimer->suivant ;

        noeudAvant->suivant = noeudApres;
        noeudApres->precedent = noeudAvant;

        FREE(noeudASupprimer);
    }

    return laListe;
}

void afficheListe(liste laListe){
    liste temp = laListe;
    printf("[");
    while(temp->suivant != NULL){
        printf("%d, ", temp->valeur);
        temp = temp->suivant;
    }
    printf("%d]\n", temp->valeur);
    return ;
}


//* Fonctions d'aide.
liste creerListeN(Nat nombreElt){
    liste maListe = listeNouv();

    Nat i = 0;
    while(i < nombreElt){
        Nat randNumber = rand() % 100 + 1;
        maListe = adjonctionEntete(maListe, randNumber) ;
        i++;
    }

    return maListe;
}

//-------------------------------------------------------------------------

Iter init(liste l){
    Iter itterateur = MALLOC(strIter);

    if( itterateur == NULL){
        fprintf(stderr, "Erreur : Allocation échouée.");
        exit(EXIT_FAILURE);
    }

    itterateur->first = l;
    itterateur->cur = l;
    if(itterateur->cur->suivant == NULL){
        itterateur->fin = vrai;
    }

    return itterateur;
}

Bool end(Iter i){
    return i->fin ;
}

Iter next(Iter i){
    if(!end(i)){
        i->cur = i->cur->suivant;
        if(i->cur == NULL){
            i->fin = vrai;
        }
    }
    
    return i;
}

Iter prev(Iter i){
    //* Le Premier noeud n'a pas de précédent.
    if(i->first != i->cur){
        i->cur = i->cur->precedent;
    }
    return i;
}

Nat value(Iter i){
    return i->cur->valeur;
}

void cherche(liste laListe, Nat valeurAChercher){
    Iter i;
    for(i = init(laListe); !end(i); next(i)){
        if(value(i) == valeurAChercher){
            printf("Valeur trouvée. \n");
            return;
        }
    }
    printf("Valeur non trouvée.\n");
    return;
}