INF_POINT = None
"""
    source : https://github.com/jacksoninfosec/elliptic-curves/blob/main/elliptic_curve.py
    Les autres fonctions je les ai écrit
"""

class EllipticCurve:
    def __init__(self, a, b, p):
        self.a = a
        self.b = b
        self.p = p

    def addition(self, P1, P2):
        if P1 == INF_POINT:
            return P2
        if P2 == INF_POINT:
            return P1

        x1 = P1[0]
        y1 = P1[1]
        x2 = P2[0]
        y2 = P2[1]

        if self.equalModp(x1, x2) and self.equalModp(y1, -y2):
            return INF_POINT

        if self.equalModp(x1, x2) and self.equalModp(y1, y2):
            u = self.reduceModp((3 * x1 * x1 + self.a) *
                                self.inverseModp(2 * y1))
        else:
            u = self.reduceModp((y1 - y2) * self.inverseModp(x1 - x2))

        v = self.reduceModp(y1 - u * x1)
        x3 = self.reduceModp(u * u - x1 - x2)
        y3 = self.reduceModp(-u * x3 - v)

        return (x3, y3)

    # helper functions

    def reduceModp(self, x):
        return x % self.p

    def equalModp(self, x, y):
        return self.reduceModp(x - y) == 0

    def inverseModp(self, x):
        for y in range(self.p):
            if self.equalModp(x * y, 1):
                return y
        return None

    # Trouve a telque aG = (x_aG, y_aG)
    def findScalar(self, G, x_aG, y_aG):
        found = False
        xTmp, yTmp = self.addition(G, G)
        a = 2  # Car la ligne precedente double le point G
        while(not found):
            if(xTmp == x_aG and yTmp == y_aG):
                found = True
            else:
                # Le point O (Point infini trouve)
                if(self.addition((xTmp, yTmp), G) == None):
                    found = True
                else:
                    xTmp, yTmp = self.addition((xTmp, yTmp), G)
                    a += 1

        return a

    def findOrder(self, G):
        found = False
        xTmp, yTmp = self.addition(G, G)
        order = 2  # Car la ligne precedente double le point G
        while(not found):
            if(self.addition((xTmp, yTmp), G) == INF_POINT):
                # Le point O (Point infini trouve)
                order+= 1
                found = True
            else:
                xTmp, yTmp = self.addition((xTmp, yTmp), G)
                order += 1

        return order

    def multiplyByScalar(self, scalar, P):
        xTmp, yTmp = self.addition(P, P)
        count = 2
        while(count < scalar):
            xTmp, yTmp = self.addition((xTmp, yTmp), P)
            count += 1

        return xTmp, yTmp

    def pointOnCurve(self, x, y):
        yAuCarre = y**2
        equation = x**3 + self.a*x + self.b

        return self.equalModp(yAuCarre, equation)
