# -*- coding:UTF-8
from Crypto.Cipher import AES
import binascii
import hashlib
from elipticCurve import EllipticCurve

"""
    Fonction de dechiffrement en utilisant l'algorithme de RSA
    Paramètres :
        cryptedMessage :
            Une liste de nombre, chaque nombre est une 
            suite de trois octets representant chacun un caractère.
        d : 
            La clé privée
        n : Le nombre n
    Retourne :
        Le message dechiffré (Une chaine de caractère). 
"""


def rsaDecrypt(cryptedMessage, d, n):
    # Une liste contenant les caractères dechiffrés.
    decryptedMessage = []
    for threeBytes in cryptedMessage:
        # Utilisation de la formule ( M = M'^d mod n)
        decryptedNumber = pow(threeBytes, d, n)
        # Decomposition en trois octet.
        threeBytesDecrypted = decryptedNumber.to_bytes(3, 'big')
        # Ajout des lettres dechiffrés dans la liste
        decryptedMessage.extend(chr(letter) for letter in threeBytesDecrypted)
    return ''.join(decryptedMessage)


ficheACChiffre = [12379464, 221568478, 5463765, 220570841, 192279664,
                  93906703, 15836803, 218383702, 107041986, 73574973, 90776304,
                  198379713, 101334295, 201988306, 42190473, 119844755, 161748843,
                  159780974, 34766228, 221568478, 112892606, 46153212, 34317921,
                  23500344, 141504156, 158440706, 170017047, 5804276, 130841495,
                  139191638, 88287086, 106977614]

e_AC = 1283
n_AC = 237922819
print(rsaDecrypt(ficheACChiffre, e_AC, n_AC))
print("\n")
"""
    Bob est bien celui qu'il pretend etre, et
    sa cle publique est (e,n)=(1117,2865761027) en decimal
    e_B = 1117
    n_B = 2865761027
"""
e_B = 1117
n_B = 2865761027
"""
    Clé privée de Bob :
        Casser e_n c'est-à-dire trouver p, q deux nombres premiers tels que
        n_B = p * q
        calculer phi(n_B) = (p-1)*(q-1),
        Resoudre l'équation e*d = 1 mod phi(n) en utilisant l'algorithme 
        d'euclide etendu.
"""
"""
    Cette fonction permet de trouver les tous facteurs premiers d'un
    nombre N
    source : stackOverFlow

"""


def calculatePrimeFactors(N):
    prime_factors = set()
    if N % 2 == 0:
        prime_factors.add(2)
    while N % 2 == 0:
        N = N // 2
        if N == 1:
            return prime_factors
    for factor in range(3, N + 1, 2):
        if N % factor == 0:
            prime_factors.add(factor)
            while N % factor == 0:
                N = N // factor
                if N == 1:
                    return prime_factors


"""
    Indicatrice d'Euleur
"""


def phi(N):
    # On casse le nombre N
    p, q = calculatePrimeFactors(N)
    return (p-1)*(q-1)


# n_B = 2865761027
phiDeN = phi(n_B)
print(f"phi({n_B}) = {phiDeN}")
# phiDeN = 2865638640
print("\n")

"""
    On utilise l'algorithme d'euclide pour trouver les coefficients de bezout
    source : cours
"""


def euclideEtendu(a, b):
    # rend r,u,v avec r=pgcd(a,b) et au+bv=r
    r0, u0, v0 = a, 1, 0
    r1, u1, v1 = b, 0, 1
    while r1 != 0:
        q = r0//r1
        r0, r1 = r1, r0-q*r1
        u0, u1 = u1, u0-q*u1
        v0, v1 = v1, v0-q*v1

    return r0, u0, v0


r, d_B, k = euclideEtendu(e_B, phiDeN)
print(f"r = {r}, d = {d_B}, k = {k}")
print("\n")
if((e_B * d_B) % phiDeN):
    print(f"Donc, la clée privee de Bob est : {d_B}")
print("\n")


"""
    Courbe eliptique choisie par Bob:
        E : y² = x^3 + 8x + 9, mod p = 37951
        forme y² = x^3 + sx + t mod p
        Par identification :
        s = 8
        t = 9
        p = 37951
"""


def equation(x, s, t):
    return x*x*x + s*x + t


p = 37951
s = 8
t = 9
# D'après la formule du cours
cardE = (p + 1) - t
# La formule ne savère pas donné la bonne valeur du coup, j'ai utilisé ce site
# https://andrea.corbellini.name/ecc/interactive/modk-add.html
# Navigateur recommandé : Google chrome
print(f"Le nombre de point de la courbe elliptique est {cardE}")

"""
    Calcul de l'ordonee du point G
"""


def calcul_y(x, p):
    yCarre = equation(x, s, t)
    # print(yCarre)
    y = 1
    while(((y * y) % p) != yCarre):
        y = y + 1
    return y


print("\n")
xG = 5
print(
    f"Le point G a pour couple (xG, yG) = ({xG}, {calcul_y(xG, p)}) mod {p}.")
yG = 2441
# Est-ce que yG = 2441 app [0, p/2] mod p = [0, 18975,5[ mod 37951
# Réponse oui, donc yG est bien 2441.

# Verifions juste si aG = (11893, 23880) appartient bien à la courbe
print("\n")
x_aG = 11893
y_aG = 23880
if(equation(x_aG, s, t) % p == y_aG**2 % p):
    print(f"aG = ({x_aG}, {y_aG}) appartient bien à la courbe")


"""
    Scalar Multiplication
    Les methodes findScalar, findOrder, multiplyByScalar prennent beaucoup de temps
    avant de s'executer c'est je les ai commenté, pour voir le résultat il faut les
    decommenter
"""
# Extended Euclidean algorithm
p = 37951
ec = EllipticCurve(8, 9, p)
G = (5, 2441)
x_aG = 11893
y_aG = 23880
# print(ec.findScalar(G, x_aG, y_aG))
a = 1320
# print(ec.findOrder(G))
order = 19064

# Clé ECDH d'Alice bP = (1990, 19625)
print("\n")
bP = (1990, 19625)
# print(ec.multiplyByScalar(a, bP))

# Clé commune
abP = (22587, 36350)
x_abP, y_abP = abP

# Juste un test d'appartenance
if(ec.pointOnCurve(x_abP, y_abP)):
    print(f"abP = {abP} est sur la courbe")
else:
    print("Le point n'est pas sur la courbe")

"""
    AES :
        la clé en hexadecimal est :
        00 00 00 00 00 00 00 00 00 00 00 00 00 00 58 3B 
    Source : chatGPT
"""
# Message chiffré
encryptedMessage = binascii.unhexlify('857344ad75709476d7aa9a5d9668bb97ba31255ae325e7c4fb015991b9f7c98b73b9d1057e7050c2b23322f3d91422628d7630bf9637a0906fadcbc9c176f236af013f3c901cedfc074f5104e1ac5d72')
print(f"Message chiffré de Bob:\n  {encryptedMessage}\n")

# Clée commune
key = x_abP.to_bytes(16, byteorder='big', signed=False)
print(f"key = {key}\n")

# Vecteur d'initialisation
IV = 00000000000000000000000000000000

# Conversion du vecteur d'initialisation en 16 octet (128 bits)
IV = IV.to_bytes(16, byteorder='big', signed=False)

# Création de l'objet cipher
cipher = AES.new(key, AES.MODE_CBC, IV)

# Dechiffrement du message
decrypted_message = cipher.decrypt(encryptedMessage)

# Conversion de chaque octet du message chiffré en caractère
plainTextMessage = ''.join(map(chr, decrypted_message))

print(f"Message dechiffré : {plainTextMessage}\n")

# Chiffrement du message alteré de Bob
cipher = AES.new(key, AES.MODE_CBC, IV)

# Message alteré de Bob en octets
messageAltered = b'Rendez-vous le 1/1/2022 a 15h00 sur le parvis de la cathedrale de Strasbourg'
print(f"Message alteré de Bob :  \n {messageAltered}\n")

# On fait en sorte d'avoir un multiple de 16 en rajoutant des octets équivalents 
# au nombre d'octets manquant
length = 16 - (len(messageAltered) % 16)
messageAltered += bytes([0x00])*length


print(f" Message alteré de Bob avec les octets manquant:  \n {messageAltered}\n")

# Chiffrement du message alteré
messageAlteredEncrypted = cipher.encrypt(messageAltered)
print(f"Message alteré de Bob chiffré :  \n {messageAlteredEncrypted}\n")


print(f"Message alteré de Bob chiffré et converti en Hex :  \n {binascii.hexlify(messageAlteredEncrypted)}")
# print(''.join(messageAlteredEncrypted))
"""
Signature de ce message : le hash (SHA256) du message altéré clair est découpé en 
blocs de 3 octets (le dernier bloc, qui n'a que deux octets, est complété par un 
octet nul à la fin), et on applique RSA sur chacun de ces 11 nombres. On obtient 11
autres nombres décimaux, que l'on rentrera séparés par des virgules.
"""

"""
    Cette fonction permet de decomposer un nombre en une liste de paquets de nombre 
    chacun pesant trois ocets
    source : chatGPT
"""
def splitInto3BytesChunks(n):
    chunks = []
    while n > 0:
        n, remainder = divmod(n, 2**24)
        chunks.append(remainder)
    return chunks



"""
    Cette fonction permet de chiffrer une liste de nombres pesant chacun trois
    octets.
    retourne la liste des nombres chiffrés
"""
def rsaEncrypt(clearMessage, d, n):
    encrytedMessage = []
    for threeBytes in clearMessage:
        cryptedBytes = pow(threeBytes, d, n)
        encrytedMessage.append(cryptedBytes)
    return encrytedMessage

clearMessageAltered = 'Rendez-vous le 1/1/2022 a 15h00 sur le parvis de la cathedrale de Strasbourg'

# Create the SHA256 hash object
hastObject = hashlib.sha256()

# Hash the message
hastObject.update(clearMessageAltered.encode('utf-8'))

# Get the hexadecimal representation of the hash
hexHash = hastObject.hexdigest()

# On rajoute l'octet nul à la fin
hexHash += "00"
print(f"\nle hash du message clair alteré en decimal avec un octet nul rajouté à la fin : \n  {hexHash}\n")

# Conversion en nombre decimal le hash du message clair et alteré.
decimalHash = int("0x" + hexHash , 16)
print(f"\nle hash du message clair alteré en decimal : \n  {decimalHash}\n")
 

list3Bytes = splitInto3BytesChunks(decimalHash)
list3Bytes.reverse()
print(f"le hash precedent mais subdiviser en une liste de 11 nombre pesant chacun trois octets :\n {list3Bytes}\n")

print(f"La liste du message alteré chiffrer est : \n {rsaEncrypt(list3Bytes, d_B, n_B)}")


